"""Persistent memoization using TinyDB."""
import atexit
import functools
import hashlib
from typing import Any, Callable, Dict

import pendulum
import jsonpickle
import tinydb
import tinydb_smartcache

from mariner import utils


class Cache:
    """Offline cache for search results.

    Attributes:
        path: Path to the database file.
        timeout: Defaults to 1 day. Period after which results should expire.
        size: Defaults to 10000. Maximum number of cached results.
        storage: Defaults to JSONStorage. Storage type for TinyDB.
    """

    def __init__(self,
                 *,
                 path: str = '~/.cache/mariner/cache.json',
                 timeout: int = 86400,
                 size: int = 10000,
                 storage: tinydb.storages.Storage = tinydb.storages.JSONStorage,
                ) -> None:
        self.path = utils.check_path(path)
        self.timeout = timeout
        self.size = size
        self.db = tinydb.TinyDB(self.path, storage)
        self.db.table_class = tinydb_smartcache.SmartCacheTable
        atexit.register(self._remove_expired)

    @property
    def newest(self) -> Dict:
        """Return last cached result."""
        newest = self.db.all()[-1]['value']
        return jsonpickle.decode(newest)

    def get_expiry(self) -> pendulum.Pendulum:
        """Return expiration time for cached results."""
        expiry = pendulum.now().add(seconds=self.timeout)
        return self._serialize_date(expiry)

    def get(self, key: str) -> Any:
        """Get torrent from database.

        Args:
            key: Key hash.

        Returns:
            Cached object.

        """
        entry = tinydb.Query()
        value = self.db.search(entry.key == key)
        if value:
            self.db.remove(entry.key == key)
            self.insert(key, jsonpickle.decode(value[0]['value']))
            return jsonpickle.decode(value[0]['value'])
        return None

    def insert(self, key: str, entry: Any) -> None:
        """Insert entry into cache.

        Args:
            key: Key hash of the entry to store.
            entry: Object to cache.
        """
        value = jsonpickle.encode(entry)
        self.db.insert({'key': key, 'time': self.get_expiry(), 'value': value})
        if len(self.db) > self.size:
            self._remove_oldest()

    def remove(self, key: str) -> None:
        """Delete key from cache.

        Args:
            key: Hash key to delete from the cache.
        """
        entry = tinydb.Query()
        self.db.remove(entry.key == key)

    def _remove_expired(self) -> None:
        """Remove old entries."""
        entry = tinydb.Query()
        now = self._serialize_date(pendulum.now())
        self.db.remove(entry.time < now)

    def _remove_oldest(self) -> None:
        """Remove oldest entry."""
        oldest = self.db.all()[0].doc_id
        self.db.remove(doc_ids=[oldest])

    @staticmethod
    def _serialize_date(date: pendulum.Pendulum) -> Dict[str, str]:
        """Serialize date so it can be stored in database.

        Args:
            date: Date to serialize.

        Returns:
            Serialized date.
        """
        return jsonpickle.encode(date)

    def __call__(self, function: Callable[..., Any]):
        """Decorator for caching function results.

        Args:
            function: Function to decorate.

        Returns:
            Cached function.
        """
        @functools.wraps(function)
        def wrapped(*args, **kwargs):
            """Cache function."""
            result = ''
            key_root = function.__name__ + \
                jsonpickle.encode(args) + jsonpickle.encode(kwargs)
            key = hashlib.md5(key_root.encode('utf8')).hexdigest()
            result = self.get(key)
            if not result:
                result = function(*args, **kwargs)
                self.insert(key, result)
            return result
        return wrapped



"""Utility functions for Mariner."""
import pathlib


def check_path(path: pathlib.Path) -> pathlib.Path:
    """Check if path exists. If not, create it.

    Args:
      path: Path to check.

    Returns:
      Resulting path.
    """
    # Convert to Path object if needed
    if not isinstance(path, pathlib.Path):
        path = pathlib.Path(path)
    path = path.expanduser()

    # If path points to a file, get the parent directory
    directory = path.parent if path.suffix else path

    # Create path if needed
    directory.mkdir(parents=True, exist_ok=True)

    return path



"""Entrance point for Mariner application."""
import os
import pathlib
import sys
from typing import List

from cliff import app, commandmanager

from mariner import __version__
from mariner import searchengine, utils, config


class Mariner(app.App):
    """Interactive CLI for Mariner."""

    def __init__(self):
        super().__init__(
            description=__version__.__description__,
            version=__version__.__version__,
            command_manager=commandmanager.CommandManager('mariner.cli'),
            deferred_help=True,
        )
        self.config = config.Config().load()

    def build_option_parser(self, description, version, argparse_kwargs=None):
        """Return an Argparse option parser for Mariner.

        Args:
            description: Mariner description.
            version: Mariner version.
        """
        # Override the help message for log output
        kwargs = {'conflict_handler': 'resolve'}
        parser = super().build_option_parser(description, version, kwargs)
        parser.add_argument(
            '--log-file',
            action='store',
            default=None,
            help='Specify a file to log output. Default ~/.local/share/mariner/mariner.log.',
        )
        return parser

    def configure_logging(self):
        """Create logging handlers for any log output."""
        # Set logging to file by default.
        if not self.options.log_file:
            log_dir = os.getenv('XDG_DATA_HOME', '~/.local/share/mariner')
            log_dir = utils.check_path(log_dir)
            self.options.log_file = pathlib.Path(log_dir) / 'mariner.log'
        super().configure_logging()

    def initialize_app(self, argv: List[str]) -> None:
        """Initialize the application.

        Args:
          argv: List[str]: List of command line arguments.
        """
        self.LOG.debug('Initialize Mariner')
        # Initialize plugins
        searchengine.SearchEngineManager().initialize_engines()

    def prepare_to_run_command(self, cmd) -> None:
        """Code to run before command.

        Args:
          cmd: Command to run.
        """
        self.LOG.debug('Preparing to run command %s', cmd.__class__.__name__)

    def clean_up(self, cmd, result, err) -> None:
        """Code to run after command.

        Args:
          cmd: Command that was run.
          result: Result of the command.
          err: Errors caught while running the command.
        """
        self.LOG.debug('Clean up %s', cmd.__class__.__name__)
        if err:
            self.LOG.debug('Got an error: %s', err)


def main(argv=sys.argv[1:]):  # pylint: disable=dangerous-default-value
    """Application entrance point.

    Args:
      argv: Defaults to sys.argv[1:]. Command line argument list.
    """
    mariner = Mariner()
    return mariner.run(argv)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



"""Contain CLI commands."""
import logging
import pathlib

from cliff import command, lister
import pyperclip

from mariner import downloader, searchengine
from mariner.plugins import distrowatch


class Download(lister.Lister):
    """Download torrent with given ID."""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """Add arguments to argument parser.

        Args:
            prog_name: Application name.

        Returns:
            Instance of argument parser.
        """
        parser = super().get_parser(prog_name)
        parser.add_argument('ID', nargs='+', type=int)
        return parser

    def take_action(self, parsed_args):
        """Download chosen torrent.

        Args:
            parsed_args: List of parsed arguments.

        Returns:
            List of downloaded torrents and the location where they were saved.
        """
        engine = distrowatch.Distrowatch()

        torrents = []
        for tid in parsed_args.ID:
            torrent = engine.get_torrent(tid)
            self.log.debug('tid=%s torrent=%s', tid, torrent)
            if torrent.torrent_url:
                torrents.append(torrent)
                self.log.debug('Torrent appended.')
                self.log.info(f'Downloading torrent ID {tid}.')
            else:
                self.log.warning(
                    f'No torrent for {torrent.name}. Use magnet link instead.')

        filelist = ((t.torrent_url, t.filename) for t in torrents)
        path = self.app.config['download_path']
        self.log.debug('filelist=%s download_path=%s', filelist, path)
        torrent_downloader = downloader.Downloader(download_path=path)
        torrent_downloader.download(filelist)

        headers = ('ID', 'Name', 'Saved to')
        columns = ((t.tid, t.name[:60], pathlib.Path(path) / t.filename)
                   for t in torrents)
        return (headers, columns)


class Magnet(command.Command):
    """Copy magnet link with given ID to clipboard."""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """Add arguments to argument parser.

        Args:
            prog_name: Application name.

        Returns:
            Instance of argument parser.
        """
        parser = super(Magnet, self).get_parser(prog_name)
        parser.add_argument('ID', nargs=1, type=int)
        return parser

    def take_action(self, parsed_args):
        """Copy chosen magnet link to clipboard.

        Args:
            parsed_args: List of parsed arguments.
        """
        engine = distrowatch.Distrowatch()
        tid = parsed_args.ID[0]
        torrent = engine.get_torrent(tid)
        self.log.debug('tid=%s torrent=%s', tid, torrent)
        if torrent.magnet_link:
            pyperclip.copy(torrent.magnet_link)
            self.log.info(f'Copied {torrent.name} magnet link to clipboard.')
            self.log.debug('magnet=%s', torrent.magnet_link)
        else:
            self.log.warning(
                f'{torrent.name} has no magnet link. Download the torrent.')


class Search(lister.Lister):
    """Search for torrents."""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """Add arguments to argument parser.

        Args:
            prog_name: Application name.

        Returns:
            Instance of argument parser.
        """
        parser = super().get_parser(prog_name)
        parser.add_argument('title', nargs=1)
        parser.add_argument('--limit', '-l', nargs='?', default=10, type=int)
        parser.add_argument('--tracker', '-t', nargs='?',
                            default=self.app.config['default_tracker'])
        return parser

    def take_action(self, parsed_args):
        """Search for a torrent.

        Args:
            parsed_args: List of parsed arguments.

        Returns:
            List of torrents found.
        """
        title = parsed_args.title[0].lower()
        limit = parsed_args.limit
        tracker = parsed_args.tracker.lower()
        self.log.debug('title=%s limit=%s tracker=%s', title, limit, tracker)

        self.log.info(f'Searching for "{title}".')
        engine = searchengine.engines[tracker]()
        torrents = engine.search(title, limit)
        self.log.debug(f'torrents={torrents}')

        headers = ('ID', 'Name', 'Available as')
        columns = ((t.tid, t.name[:80], t.mods) for t in torrents)
        return (headers, columns)



"""Module for searching torrents on PirateBay."""
from typing import List, Tuple

import bs4

from mariner import searchengine

Magnet = str
Name = str
Url = str


class PirateBay(searchengine.SearchEngine):
    """Represents PirateBay search engine."""

    search_url = 'https://thepiratebay.org/search/'

    @staticmethod
    def _parse(raw: str) -> List[Tuple[Name, Magnet, Url]]:
        """Parse result page.

        Args:
          raw: Raw HTML results page to parse.

        Returns:
            List of torrent names with magnet links.
        """
        soup = bs4.BeautifulSoup(raw, 'lxml')
        content = soup.find('table', id='searchResult')
        try:
            torrents = content.find_all('tr')
        except AttributeError:
            # No search result
            return []
        else:
            results = []
            for torrent_ in torrents[1:]:
                raw_name = torrent_.find('a', class_='detLink')
                name = PirateBay._parse_name(raw_name)

                links = torrent_.find_all('a')
                magnet = links[3].get('href')
                # PirateBay has only magnet links
                url = None

                results.append((name, magnet, url))
            return results

    @staticmethod
    def _parse_name(raw: str) -> Name:
        """Parse torrent name.

        Args:
          raw: Raw name to parse.

        Returns:
            Parsed name.
        """
        name = raw.string
        if name is None:
            name = raw['title'].split(" ")[2:]
            name = ' '.join(str(x) for x in name)
        return name.strip()
